/*
 * single linked list
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "list.h"
#include "xmalloc.h"

struct slist *slist_new()
{
	struct slist *s;

	s = xmalloc0(sizeof(*s));
	s->data = NULL;
	s->next = NULL;

	return s;
}

void slist_free(struct slist **s)
{
	if (!s || *s == NULL)
		xerror(1, "slist_free: s NULL");

	xfree(*s);
	*s = NULL;
}

void slist_free_all(struct slist **slist, void (*release)(void *data))
{
	struct slist *iter, *next;

	if (!slist || *slist == NULL)
		xerror(1, "slist_free_all: slist NULL");

	iter = *slist;
	while (iter) {
		next = iter->next;
		if (release != NULL)
			release(iter->data);
		xfree(iter);
		iter = next;
	}
	*slist = NULL;
}

struct slist *slist_add(struct slist *slist, struct slist *e)
{
	if (!e)
		xerror(1, "slist_add: e NULL");

	if (e->next != NULL)
		xwarning("slist_add: element in list");

	if (slist)
		e->next = slist;
	slist = e;

	return slist;
}

struct slist *slist_add_tail(struct slist *slist, struct slist *e)
{
	struct slist *iter;

	if (!e)
		xerror(1, "slist_add_tail: e NULL");

	if (e->next != NULL)
		xwarning("slist_add_tail: element in list");

	if (slist) {
		for (iter = slist; iter->next != NULL; iter = iter->next)
			;
		iter->next = e;
		//e->next = NULL;
	} else {
		slist = e;
		//e->next = NULL;
	}

	return slist;
}

struct slist *slist_extract(struct slist **slist)
{
	struct slist *s = NULL;

	if (!slist)
		xerror(1, "slist_remove: slist NULL");

	if (*slist) {
		s = *slist;
		*slist = (*slist)->next;
		s->next = NULL;
	}

	return s;
}

struct slist *slist_unlink(struct slist *slist, struct slist *s)
{
	struct slist *iter, *prev = NULL;

	iter = slist;
	while (iter) {
		if (iter == s) {
			if (prev)
				prev->next = iter->next;
			else
				slist = iter->next;
			iter->next = NULL;
			return slist;
		}
		prev = iter;
		iter = iter->next;
	}

	return slist;
}

struct slist *slist_reverse(struct slist *slist)
{
	struct slist *iter, *reverse = NULL;

	iter = slist;
	while (iter) {
		slist = iter->next;
		if (reverse)
			iter->next = reverse;
		else
			iter->next = NULL;
		reverse = iter;
		iter = slist;
	}

	return reverse;
}

size_t slist_count(struct slist *slist)
{
	size_t count = 0;
	struct slist *iter;

	iter = slist;
	while (iter) {
		iter = iter->next;
		count++;
	}

	return count;
}

struct slist *slist_find(struct slist *slist, const void *data,
		int (*compare)(const void *a, const void *b))
{
	struct slist *iter;

	for (iter = slist; iter != NULL; iter = iter->next) {
		if (compare) {
			if ((*compare)(iter->data, data) == 0)
				return iter;
		} else {
			if (iter->data == data)
				return iter;
		}
	}

	return NULL;
}

void slist_traverse(struct slist *slist, slistcallback_t callback, void *opaque)
{
	int res;
	struct slist *iter;

	for (iter = slist; iter != NULL; iter = iter->next) {

		res = callback(iter->data, opaque);

		if (res)
			return;
	}
}

